// Color depth has to be 16 bits if DMA is used to render image
#define COLOR_DEPTH 16

// 128x128 for a 16 bit colour Sprite (32Kbytes RAM)
// Maximum is 181x181 (64Kbytes) for DMA -  restricted by processor design
#define IWIDTH 160
#define IHEIGHT 128

// Size of cube image
// 358 is max for 128x128 sprite, too big and pixel trails are drawn...
#define CUBE_SIZE 358

#include <Button.h>
#include <TFT_eSPI.h>
#include "cellmap.hpp"
#include "bluetooth.hpp"

// Library instance
TFT_eSPI tft = TFT_eSPI();  // Declare object "tft"

// Used for fps measuring
uint16_t counter = 0;
long startMillis = millis();
uint16_t interval = 100;

// Frames per second
String fps = "0fps";

// IO Objects
Button start_game_button(0, 100);

// Create two sprites for a DMA toggle buffer
TFT_eSprite spr[2] = {TFT_eSprite(&tft), TFT_eSprite(&tft) };

const uint8_t SCALE = 2;
CellMap current_map(IWIDTH/SCALE , IHEIGHT/SCALE, SCALE, &spr[0]);
CellMap buffer_map(IWIDTH/SCALE, IHEIGHT/SCALE, SCALE, &spr[1]);
uint16_t NUM_INITIAL_CELLS = 50;
bool game_started = false;

uint8_t cursor_x, cursor_y;
bool draw_cursor = false;

void reset_game();
void start_game();
void setChangedHandler();

char characteristic_data = 'H';

/////////////////////////////////////////////////////////// setup
////////////////////////////////////////////////////////////
void setup() {
  randomSeed(analogRead(26));

  Serial.begin(115200);
  start_game_button.begin();
  tft.init();
  tft.fillScreen(TFT_BLACK);

  // Define cprite colour depth
  spr[0].setColorDepth(COLOR_DEPTH);
  spr[1].setColorDepth(COLOR_DEPTH);

  // Create the 2 sprites
  spr[0].createSprite(IHEIGHT, IWIDTH);
  spr[1].createSprite(IHEIGHT, IWIDTH);

  // Define text datum and text colour for Sprites
  spr[0].setTextColor(TFT_BLACK);
  spr[0].setTextDatum(MC_DATUM);
  spr[1].setTextColor(TFT_BLACK);
  spr[1].setTextDatum(MC_DATUM);

  tft.initDMA(); // Initialise the DMA engine (tested with STM32F446 and
  // STM32F767)

  // Animation control timer
  startMillis = millis();

    // set callbacks
  BTstack.setBLEDeviceConnectedCallback(deviceConnectedCallback);
  BTstack.setBLEDeviceDisconnectedCallback(deviceDisconnectedCallback);
  BTstack.setGATTCharacteristicRead(gattReadCallback);
  BTstack.setGATTCharacteristicWrite(gattWriteCallback);

  // setup GATT Database
  BTstack.addGATTService(new UUID("B8E06067-62AD-41BA-9231-206AE80AB551"));
  BTstack.addGATTCharacteristic(new UUID("f897177b-aee8-4767-8ecc-cc694fd5fcef"), ATT_PROPERTY_READ, "This is a String!");
  BTstack.addGATTCharacteristicDynamic(new UUID("f897177b-aee8-4767-8ecc-cc694fd5fce0"), ATT_PROPERTY_READ | ATT_PROPERTY_WRITE | ATT_PROPERTY_NOTIFY, 0);

  // startup Bluetooth and activate advertisements
  BTstack.setup();
  BTstack.startAdvertising();
}

/////////////////////////////////////////////////////////// loop
//////////////////////////////////////////////////////
void loop() {
  uint32_t updateTime = 0;  // time for next update
  int wait = 33; // 30 FPS

  // Grab exclusive use of the SPI bus
  tft.startWrite();

  // Loop forever
  while (1) {

    // Serial.println("Bruh moment");
    if (!game_started && start_game_button.pressed() || characteristic_data == 'C') {
      reset_game();
      start_game();
      characteristic_data = 'N';
    } else if (game_started && start_game_button.pressed()) {
      reset_game();
    }


    if (updateTime <= millis()) {
      updateTime = millis() + wait;
      
      if (game_started) {
        current_map.next_generation(buffer_map);
        current_map.copy_cells(buffer_map);

        tft.pushImageDMA(0, 0, IHEIGHT, IWIDTH, (uint16_t*)current_map.tft->getPointer());
      }

      counter++;
      // only calculate the fps every <interval> iterations.
      if (counter % interval == 0) {
        long millisSinceUpdate = millis() - startMillis;
        fps = String((int)(interval * 1000.0 / (millisSinceUpdate))) + " fps";
        // Serial.println(fps);
        startMillis = millis();
      }
    }
  }  // End of forever loop

  // Release exclusive use of SPI bus ( here as a reminder... forever loop
  // prevents execution) 
  tft.endWrite();
}

void reset_game() {
  tft.fillScreen(TFT_BLACK);
  current_map.reset();
  buffer_map.reset();

  game_started = false;
}

void start_game() {
  for (uint16_t i; i < NUM_INITIAL_CELLS; i++) {
    uint8_t x, y;
    x = random(IHEIGHT/SCALE/4);
    y = random(IWIDTH/SCALE/4);
    buffer_map.set_cell(x + (IHEIGHT/SCALE)/2, y + (IWIDTH/SCALE)/2);
  }
  current_map.copy_cells(buffer_map);

  game_started = true;
}
/*
  Original licence:
  The MIT License (MIT)
  Copyright (c) 2017 by Daniel Eichhorn
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
