#include "cellmap.hpp"

CellMap::CellMap(uint8_t h, uint8_t w, uint8_t s, TFT_eSprite* tft_instance) {
  tft = tft_instance;
  scale = s;

  width = w;
  byte_width = ((w + 7) / 8) + 2;  //+2 for padding on both side.

  height = h;
  byte_total_length = byte_width * (h + 2);  //+2 for padding on both side.

  cells = new uint8_t[byte_total_length];
  memset(cells, 0, byte_total_length);
}

void CellMap::reset() {
  memset(cells, 0, byte_total_length);
  tft->fillSprite(TFT_BLACK);
}

void CellMap::copy_cells(CellMap& sourcemap) {
  unsigned char* cell_ptr;
  int i;

  // Copy left and right edges into padding bytes on right and left
  cell_ptr = sourcemap.cells + byte_width;
  for (i = 0; i < height; i++) {
    *cell_ptr = *(cell_ptr + byte_width - 2);  // Copy left edge from right edge
    *(cell_ptr + byte_width - 1) =
        *(cell_ptr + 1);     // Copy right edge from left edge
    cell_ptr += byte_width;  // Advance to next row
  }
  // Copy top and bottom edges into padding bytes on bottom and top
  memcpy(sourcemap.cells,
         sourcemap.cells + byte_total_length - (byte_width * 2), byte_width);
  memcpy(sourcemap.cells + byte_total_length - byte_width,
         sourcemap.cells + byte_width, byte_width);

  // Copy all cells to the destination
  memcpy(cells, sourcemap.cells, byte_total_length);
}

void CellMap::set_cell(uint8_t x, uint8_t y) {
  // Points to start of a byte element
  uint8_t* cell_ptr = cells + ((y + 1) * byte_width) + ((x / 8) + 1);

  // Last 4 bits of x gives offset from start of byte
  *(cell_ptr) |= 0x80 >> (x & 0x07);
}

void CellMap::clear_cell(uint8_t x, uint8_t y) {
  // Points to start of a byte element
  uint8_t* cell_ptr = cells + ((y + 1) * byte_width) + ((x / 8) + 1);

  // Last 4 bits of x gives offset from start of byte
  *(cell_ptr) &= ~(0x80 >> (x & 0x07));
}

uint8_t CellMap::cell_state(uint8_t x, uint8_t y) {
  // Points to start of a byte element
  uint8_t* cell_ptr = cells + ((y + 1) * byte_width) + ((x / 8) + 1);

  return (*cell_ptr & (0x80 >> (x & 0x07))) ? 1 : 0;
}

uint8_t CellMap::count_neighbours(uint8_t x, uint8_t y) {
  uint8_t* cell_ptr;
  uint8_t mask, neighbour_count;

  // Start pointer on upper left neighbour cell
  // Small optimization from (x-1)/8 + 1 to (x+7)/8
  cell_ptr = cells + (y * byte_width) + ((x + 7) / 8);

  // Bitmask for a bit that is to the left of the selected cell in its byte
  // chunk
  mask = 0x80 >> ((x - 1) & 0x07);

  // ----------------------------------Left-------------------------------------
  // Upper left
  neighbour_count = (*cell_ptr & mask) ? 1 : 0;
  if (neighbour_count) {
    Serial.println("Upper left neighbour found");
  }
  // Left
  if (*(cell_ptr + byte_width) & mask) {
    Serial.println("Left neighbour found");
    neighbour_count++;
  }
  // Lower left
  if (*(cell_ptr + (byte_width * 2)) & mask) {
    Serial.println("Lower left neighbour found");
    neighbour_count++;
  }

  // ----------------------------------Upper------------------------------------
  // Handles if cell is at the end of its byte chunk
  // >>= sets mask to itself shifted to the right, now we point to the upper
  // neighbour
  if ((mask >>= 1) == 0) {
    mask = 0x80;  // Now masks first cell in its byte chunk
    cell_ptr++;  // Move to the next byte chunk so mask points to the upper cell
    if (neighbour_count) {
      Serial.println("Cell is at the end of its byte chunk");
    }
  }
  // Upper
  if (*cell_ptr & mask) {
    neighbour_count++;
    Serial.println("Upper neighbour found");
  }
  // Lower
  if (*(cell_ptr + (byte_width * 2)) & mask) {
    neighbour_count++;
    Serial.println("Lower neighbour found");
  }

  // ----------------------------------Right------------------------------------
  // Do the same, now points to right neighbour
  if ((mask >>= 1) == 0) {
    mask = 0x80;  // Now masks first cell in its byte chunk
    cell_ptr++;  // Move to the next byte chunk so mask points to the upper cell
    if (neighbour_count) {
      Serial.println("Cell is at the end of its byte chunk");
    }
  }

  // Upper right
  if (*cell_ptr & mask) {
    Serial.println("Upper right neighbour found");
    neighbour_count++;
  }
  // Right
  if (*(cell_ptr + byte_width) & mask) {
    Serial.println("Right neighbour found");
    neighbour_count++;
  }
  // Lower right
  if (*(cell_ptr + (byte_width * 2)) & mask) {
    Serial.println("Lower right neighbour found");
    neighbour_count++;
  }

  return neighbour_count;
}

void CellMap::next_generation(CellMap& dest) {
  uint8_t x, y, neighbour_count;

  uint8_t byte_widthX2 = byte_width << 1;
  uint8_t *cell_ptr, *current_cell_ptr, mask, current_mask;
  uint8_t *base_cell_ptr, *row_cell_ptr, base_mask;
  uint8_t* dest_cell_ptr = dest.cells;

  // Process all cells in the current cellmap
  row_cell_ptr = cells;           // point to upper left neighbor of
                                  // first cell in cell map

  for (y = 0; y < height; y++) {  // repeat for each row of cells

    // Cell pointer and cell bit mask for first cell in row
    base_cell_ptr = row_cell_ptr;  // to access upper left neighbor
    base_mask = 0x01;              // of first cell in row

    for (x = 0; x < width; x++) {  // repeat for each cell in row
    
      // First, count neighbors
      // Point to upper left neighbor of current cell
      cell_ptr = base_cell_ptr;  // pointer and bit mask for
      mask = base_mask;          // upper left neighbor
      // Count upper left neighbor
      neighbour_count = (*cell_ptr & mask) ? 1 : 0;
      // Count left neighbor
      if ((*(cell_ptr + byte_width) & mask))
        neighbour_count++;
      // Count lower left neighbor
      if ((*(cell_ptr + byte_widthX2) & mask))
        neighbour_count++;
      // Point to upper neighbor
      if ((mask >>= 1) == 0) {
        mask = 0x80;
        cell_ptr++;
      }
      // Remember where to find the current cell
      current_cell_ptr = cell_ptr + byte_width;
      current_mask = mask;
      // Count upper neighbor
      if ((*cell_ptr & mask))
        neighbour_count++;
      // Count lower neighbor
      if ((*(cell_ptr + byte_widthX2) & mask))
        neighbour_count++;
      // Point to upper right neighbor
      if ((mask >>= 1) == 0) {
        mask = 0x80;
        cell_ptr++;
      }
      // Count upper right neighbor
      if ((*cell_ptr & mask))
        neighbour_count++;
      // Count right neighbor
      if ((*(cell_ptr + byte_width) & mask))
        neighbour_count++;
      // Count lower right neighbor
      if ((*(cell_ptr + byte_widthX2) & mask))
        neighbour_count++;
      if (*current_cell_ptr & current_mask) {
        if ((neighbour_count != 2) && (neighbour_count != 3)) {
          *(dest_cell_ptr + (current_cell_ptr - cells)) &=
              ~current_mask;  // turn off cell
          tft->fillRect(x * scale, y * scale, scale, scale, TFT_BLACK);
        }
      } else {
        if (neighbour_count == 3) {
          *(dest_cell_ptr + (current_cell_ptr - cells)) |=
              current_mask;  // turn on cell
          tft->fillRect(x * scale, y * scale, scale, scale, TFT_WHITE);
        }
      }
      // Advance to the next cell on row
      if ((base_mask >>= 1) == 0) {
        base_mask = 0x80;
        base_cell_ptr++;  // advance to the next cell byte
      }
    }
    row_cell_ptr += byte_width;  // point to start of next row
  }
}
